import hashlib
import sys 
import math

# sha256 does not seem vulnerable
# sha256 uses 32-byte blocks, but we only take 8 bytes (1/4)

# vulnerability: ECB mode. can split cipher up into 3 16-byte blocks and crack individually
# inputs are blocks of 16 bytes
# cipher blocks:
# - 15e1c8c4bd2fcb12 838d3f48e8fc567a
# - df4fac1c36648eb2 70ad899a717dfadf
# - ccd9b49fc32a7267 11dd1d7faab7ed81

# we can crack the 8 byte half-blocks, but cracking speed is in magnitude of 25M/s for full SHA256 on GPU
# we would need to crack 300 blocks

# the last block possibly has a length of less than 16 (eg. something short like 3 or 4).
# we could try to crack this block, but it would only give us part of the plaintext
# tried 0-4 characters before the }, but didnt find anything

# we also know part of the plaintext, ie. it starts with RTN{ and ends with }
# also, the input will be ascii text (!!!), which drastically reduces the search keyspace FOR THE FULL BLOCK (not for half-blocks!!)
# search keyspace would be abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_ (63 chars)
# this would mean worst case search space is 63^16 (still way too much)



BLOCK_SIZE = 16

def encrypt(plain_text, key):
    result = bytearray()

    padding_size = math.ceil(len(plain_text) / BLOCK_SIZE) * BLOCK_SIZE

    plain_text = plain_text.ljust(padding_size, b'\0') # add 00 bytes to the right, up to multiple of 16 BYTES (32 octets)

    for i in range(0, len(plain_text), BLOCK_SIZE):
        result.extend(encrypt_block(plain_text[i:i+BLOCK_SIZE], key))

    return result


def encrypt_block(block, key):
    print(f"encrypt_block {block.hex()}")
    a = block[0:BLOCK_SIZE // 2]
    b = block[BLOCK_SIZE // 2:BLOCK_SIZE]

    for k in key:
        print("\tk[i]")

        # take quarter of hash!
        c = hashlib.sha256(b + bytes([k])).digest()[0:BLOCK_SIZE // 2]

        print(f"\t\tc = {c.hex()} sha256(prev_b + key[i]) -> sha256({b.hex()} + {bytes([k]).hex()})")

        # a ^= sha256(b+key[i])
        a = bytes(x ^ y for x, y in zip(c, a))

        # swap(a, b)
        b, a = a, b

        print(f"\t\ta = {a.hex()} (prev_b)")
        print(f"\t\tb = {b.hex()} (c ^ prev_a)")

    return b + a

if __name__ == "__main__":
    # if len(sys.argv) <= 1:
    #     print(f"Usage {sys.argv[0]} <plain_text> <public_key>")
    #     sys.exit(0)
    
    plain_input = "0000000000000000FFFFFFFFFFFFFFFF" # sys.argv[1]
    # public_input = "537472347762337272795f536d303074686933535f42337374" # sys.argv[2]
    public_input = "000102" # sys.argv[2]

    # both types are "bytes"
    p = bytes.fromhex(plain_input) # plain
    k = bytes.fromhex(public_input) # public

    encrypted = encrypt(p, k)
    print(encrypted.hex())

# Cipher text: 15e1c8c4bd2fcb12838d3f48e8fc567adf4fac1c36648eb270ad899a717dfadfccd9b49fc32a726711dd1d7faab7ed81
# ciphertext is 48 bytes, 3 blocks, so cleartext should be between 33 and 48 bytes
# Public key: 537472347762337272795f536d303074686933535f42337374
# pubkey is 50 chars
